#ifndef SF_FILEMANAGER
#define SF_FILEMANAGER

using namespace std;

// combines the text names and allocates enough memory in pCom, takes pTex1's file ext
bool CombineTexNames(TCHAR*& pCom, const TCHAR* pTex1, const TCHAR* pTex2);

/**
	\class SFFileManager
   Resource Management
   The file manager will control the memory allocated to the video card.
   It will handle which objects are loaded and if any new objects are requested, it
   will first check if there are any similar existing ones. The render objects themselves
   will be saved in a database. A database will save the resource ids of the textures
   and vertex/index buffers it is using. When a filemanager loads a file, 
   it will first search its own loaded objects then go to the database objects. 
   After that it will fill the buffers, add a reference count on the object and load it to memory (if not loaded yet).
   An id of 0 for anything means that it isn't managed and that object is responsible for cleaning
   itself up after a Release call.

   Saved resource files
   Resources are going to be shared, there will be one instance of a resource.
   When the databse saves objects, it will save a list of object ids. This list will
   be in the order that the object ids come from. Beside each object id will be the
   offset into the file that its memory is present. This will be filled after the resource
   is done writing. When data is queried, it will open the file and load either 
   to video memory or system.

   A resource object will be responsible for writing and deleting
   a file or chunk of memory once it goes out of scope. This file/chunk will hold processed 
   buffer information. The file is stored in a special directory defined by the level or database if it
   is a level editor file. The game file for a file manager will have all the information of all objects and
   all small files will just be concatenated with the same format as their single counterparts.
   We are assuming that one resource has a unique filename so therefore resources can't 
   delete each other's files. During the lifetime of the editor, resources will be created 
   and destroyed all the time. The user has capability to delete a resource 
   and never see it again, the problem with this is that individual resources
   need to know where their files are (they don't have path names). In this case, the deletion
   is implemented in Release(). It will take two parameters, the first to assert the deletion
   and the other is the directory name. This is not necessary and if any of them are invalid, then
   the file won't get deleted (in the case of the actual game). The resource will also have
   Load and FilenameChanged methods. Load loads and processes a base file based on teh current
   filename that is loaded. FilenameChanged does that and changes the filename. Both delete
   their previous resources except that FilenameChanged deletes the old filename resource.
   Each resource will also have Read/WriteResData functions. These serve to load/store the resource
   into a separate file or the current file. Read/WriteHeader save and restore variables
   that will be edited by the user. Load/ReleaseVideoRes loads the resource into video
   memory, this only works if there is a system resource. This is a must if the resource is going 
   to be Rendered or is in the vicinity of the  Render area. 

   There are two types of Destroy functions: DestroyRes, DestroyAll. DestroyAll calls DestroyRes.
   Both destroy functions do not touch the memory necessary to rebuild the object from scratch,
   their ids or the save flag (if in editor). So everything that will appear in the editor's
   dialog boxes when the object is really far away or not even in the current level,
   should be left to the destructor of the object. DestroyRes destroys everything that is mesh/texture related
   like vertex buffers, vertex counts, vertex sizes, etc. This will be used whenever there is a change
   in the mesh filename, compression type, or render type (tangent space info). The DestroyAll destorys
   everything else for example in animation objects, it releases the memory held by each animation.
   DestroyAll will be used in the Release and Read calls. Note that ReadHeader will also have its
   own memory releases for the editor specific values like filenames, options, etc. In editor mode,
   the functions will reset values. These values won't be needed for game mode because a Destroy usually
   means that not only the resources are getting destroyed but the object, so the destructor follows right after.

   Since resources are going to be shared, all a render object has to do is call Release
   on the resource. Whether the resource goes out of scope or not is its business (reference count).

   Usage
   There will be one instance of the object and that it will be owned by the database.
   Once a render object has been requested, all the code that requested it needs to worry about
   is calling Release after it is done, calling LoadVideoRes whenever in range. For every call
   to LoadVideoRes there can be only one matching call to ReleaseVideoRes (it is reference counted).
   If a render object is not handled by the manager, then calling Release will delete all its memory, but
   won't deallocate its memory; therefore make sure you know which objects are handled by the manager.
   The manager does not initially load the resouces to video memory if bSystemRes is true.
*/

// tested with bitwise masks, since one resource can inherit another like anim obj
#define SFRESOURCE_TEXTURE		1
#define SFRESOURCE_OBJECT		2
#define	SFRESOURCE_ANIMOBJ		6

/// Strings used to get the input data
#define SHADERINPUT				_T("i")
#define SHADERINPUT_POS			_T("v_pos")
#define SHADERINPUT_BLEND		_T("v_blend")
#define SHADERINPUT_BINDICES	_T("v_blendindices")
#define SHADERINPUT_NORMAL		_T("v_normal")
#define SHADERINPUT_TANGENT		_T("v_tangent")
#define SHADERINPUT_TEX			_T("v_tex")			///< concat a number for indexing
#define SHADERINPUT_COLOR		_T("v_color")		///< concat a number for indexing

/// Variables that will be outputted by SFObject::WriteShaderData
#define SHADEROBJ_POS			_T("b_objpos")
#define SHADEROBJ_NORMAL		_T("b_objnorm")
#define SHADEROBJ_TANGENT		_T("b_objtangent")
#define SHADEROBJ_TEX			_T("b_objtex")			///< concat a number for indexing
#define SHADEROBJ_COLOR			_T("b_objcolor")		///< concat a number for indexing

#define SFM_SHADERCONST		0x0100		///< start of sffilemanager shader consts
#define SFM_COMPRESSMAT		(SFM_SHADERCONST+0)
#define SFM_COMPRESSMAT_VAR	_T("b_mCmpMat")

// SFObject input flags

/// regular, ordered input semantics in the way they would be expected to appear in memory
#define		RTPOS_MASK	0x0007			// mask for position data
#define		RTBLEND0	0x0000
#define		RTBLEND1	0x0001			///< 0 vertex blend weight, 1 index
#define		RTBLEND2	0x0002			///< 1 vertex blend weights, 2 indices
#define		RTBLEND3	0x0003			///< 2 vertex blend weights, 3 indices
#define		RTBLEND4	0x0004			///< 3 vertex blend weights, 4 indices
#define		RTNORMAL	0x0010
#define		RTTANGENT	0x0020
#define		RTCOLOR		0x0040

#define		RT_SETTEX(rt, num)		(rt) = ((rt)&0xf8ff)|((num)<<8)
#define		RT_GETTEX(rt)			(((rt)>>8)&7)

/// compression flags, used in the same DWORD as the RT flags
#define		COMPRESSION_MASK		0xffff0000		///< mask for compression data since shares with RT flags
#define		CTPOS					0x00010000		///< SHORT4
/// blend1 takes up just one index, so compressing it is only shooting yourself in the foot
#define		CTBLEND		0x00020000		///< the blends are stored in D3DCOLOR,
										///< if rtblend1: D3DCOLOR1.z - index
										///< rtblend2: D3DCOLOR1.x - blend; D3DCOLOR1.zw - index 12
										///< rtblend34: D3DCOLOR1.xy(z) - blends; D3DCOLOR2 - indices
#define		CTPOS24		0x00040000		///< if using CTPOS24, then CTPOS and CTNORMALS bit is set also, still SHORT4
#define		CTNORMALS	0x00100000		///< D3DDECLTYPE_DEC3N - each component is a signed 10bit integer; in shader, gets normalized
#define		CTTEXTURE	0x01000000		///< (USHORT2) the fractional part is extracted from the coords and
										///< then 65535 is multiplied to get a 16 bit unsigned value. normalized in shader.

#ifdef SHADOWFRAME_EDITOR
//// maps a name to an index
typedef map<sfstring, WORD, less< sfstring > > MAPNAMES;
#endif

namespace SFManagerSpace {
	/// returns all the registered constants that are used by sffilemanager
	void GetRegisteredConsts(list<SHADERCONST*>& listConsts);
}

class SFResource
{
	friend class SFFileManager;
	friend class SFSpaceFileManager;
	friend class SFRenderObj;

protected:
	SFResource(WORD wResType);
	virtual ~SFResource();

	virtual void DestroyAll();
	virtual void DestroyRes() {};

	virtual bool Release();					// returns true if reference count reached 0

#ifdef SHADOWFRAME_EDITOR
	inline void AddEditorRef()		{ assert( byEditorRef < 255); ++byEditorRef; }
	bool ReleaseEditorRef();
#endif

public:

	inline void AddRef()			{ assert( byRef < 255 ); ++byRef; }

	virtual HRESULT LoadVideoRes(LPD3DDEV pd3dDevice);
	virtual void ReleaseVideoRes();

#ifdef SHADOWFRAME_EDITOR
	virtual TCHAR* GetSaveExt() = 0;			///< gets the filename extension to save in
	virtual bool GetResFilename(sfstring&);		///< gets the resource filename this obj controls

	// deletes the file pointed to by pFilename first
	virtual void DeleteResFile(TCHAR* pCurDirectory);

	/// called when the resource file needs to be changed
	/// \param filename The new file of the resource
	/// \param pCurDirectory Current directory of the manager holding this resource
	/// \param pd3dDevice
	/// \param dwLoadOpt One of RESOPT_,TEXOPT_,OBJOPT_, etc flags that specify options about the resource
	virtual HRESULT FilenameChanged(const TCHAR* filename, TCHAR* pCurDirectory, LPD3DDEV pd3dDevice, 
		bool bSystemMem, WORD wLoadOpt = 0, WORD wLoadOptMask = 0);

protected:
	/// releases all memory and loads a mesh from pFilename param
	virtual HRESULT Load(LPD3DDEV pd3dDevice, bool bSystemMem);

	virtual void WriteHeaderEditor(FILE* f);		// calls WriteHeader first
	virtual void ReadHeaderEditor(FILE* f);			// calls ReadHeader first

	virtual void WriteHeader(FILE* f) {}
	virtual bool WriteResData(LPD3DDEV pd3dDevice) { return true; }

	/// builds the resource file if it doesn't exist and appends it to the open file handle
	/// calls WriteResData to write the file if not written yet
	virtual bool BuildGameFile(FILE* f, LPD3DDEV pd3dDevice);

	/// If the data is saved, reads from that data using ReadResDataFile;
	/// otherwise uses LoadMesh and writes the data to a file
	virtual HRESULT ReadResData(LPD3DDEV pd3dDevice, bool bSystemMem);

public:
	inline bool	IsSaved()						{ return bSaved; }
	inline const TCHAR* GetFilename()			{ return strfile.c_str(); }
#endif

public:
	virtual void ReadHeader(FILE* f);			// destroys everything first and sets bSaved flag to true before reading anything

#ifdef SHADOWFRAME_EDITOR
protected:
#endif
	/// in editor, should only be called by ReadResData(TCHAR*, ...)
	virtual HRESULT ReadResDataFile(FILE* f, LPD3DDEV pd3dDevice, bool bSystemMem)  { return S_OK; }

public:
	virtual void WriteShaderData(sfstring& sfinput, sfstring& sfvscode) {}
	virtual void SetShaderConst(ISFEffect* peff) {}

	inline BYTE		GetRef()				{ return byRef; }
	inline BYTE		GetLoadRef()			{ return byLoadRef; }
	inline WORD		GetId()					{ return wId; }
	inline WORD		GetResType()			{ return wResType; }

protected:
	WORD wId;						///< id for map index and resource id, if not 0, then has entry in manager
	WORD wResOpt;					///< the resource options
	BYTE byRef, byLoadRef;

#ifdef SHADOWFRAME_EDITOR
	sfstring strfile;					/// pointer to a .bmp, .tga, .dds, .x, etc
	BYTE byEditorRef;					///< reference count for editor only, if goes to zero, resource is removed from mngr
	bool bSaved;						///< true if the buffer has been saved in some file,
										///< flag resets whenever pFilename changes (only valid for SAVE_MODELFILENAME)
#endif

private:
	WORD wResType;					///< the type of resource
};

// Classes that use instances of the resource objects
class SFTexture;
class SFObject;
class SFAnimObject;
class SFFileManager;

class SFRenderObj
{
	friend class SFFileManager;
	friend class SFSpaceFileManager;
	friend class SFPickupRenderObj;
	friend class SFItemFileManager;

protected:

	SFRenderObj(SFFileManager*);
	virtual ~SFRenderObj();

	virtual bool Release();

public:
	struct MATERIAL
	{
		MATERIAL() : wMtrlId(0) {}
		SHADERCONSTVEC vecConsts;
		WORD wMtrlId;
	};

	virtual void AddRef(LPD3DDEV pd3dDevice, bool bSystemMem, bool bLoad = true);

	virtual HRESULT LoadVideoRes(LPD3DDEV pd3dDevice);
	virtual void ReleaseVideoRes();

	virtual void ReadHeader(FILE* f);

#ifdef SHADOWFRAME_EDITOR
	virtual void WriteHeader(FILE* f);

	//@{
	virtual void WriteHeaderEditor(FILE* f);		///< Editor specific data. calls WriteHeader
	virtual void ReadHeaderEditor(FILE* f);
	//@}

	/// sets the resource given the index and the new resource id. Will need to load
	/// the resource in order to find out the attribute table
	virtual HRESULT SetObject(WORD wNewResId, LPD3DDEV pd3dDevice, bool bSystemMem);

	/// Sets a new material, and tries to preserve the old constants from the material.
	/// Adds and removes constants from vecpMtrls[wMtrl]->vecConsts such that the ids in vecNewConsts match.
	/// Only preserves constants if they have the same name and data type.
	virtual HRESULT SetMtrl(UINT uindex, WORD wMtrl, vector<SHADERCONST*>& vecNewConsts, LPD3DDEV pd3dDevice);

	HRESULT SetNumMaterials(UINT uNumMtrls, LPD3DDEV);

	virtual HRESULT RebuildTexIds(LPD3DDEV pd3dDevice);
	inline bool DispEffects()		{ return bDispEffect; }
#else
	inline bool DispEffects()			{ return true; }
#endif

	
	inline SFObject*	GetSFObject()		{ return pObj; }
	inline WORD			GetId()				{ return wId; }
	inline BYTE			GetRef()			{ return byRef; }
	inline BYTE			GetLoadRef()		{ return byLoadRef; }
	inline WORD			GetResId()			{ return wResId; }
	inline DWORD		GetNumMtrls()		{ return (DWORD)vecpMtrls.size(); }
	inline SFFileManager* GetMngr()			{ return pMngr; }
	inline const MATERIAL*	GetMtrl(UINT uindex)			{ assert( uindex < (UINT)vecpMtrls.size()); return vecpMtrls[uindex]; }

	sfstring strName;
	SFParticleEffectsMngr sfpeParticleMngr;		// only one effects script per render obj

protected:
	vector< MATERIAL* > vecpMtrls;		// one vector of constants for each material 
	
	SFObject* pObj;

	WORD wResId;
	WORD wId;
	BYTE byRef, byLoadRef;

	SFFileManager* pMngr;

#ifdef SHADOWFRAME_EDITOR
	bool bDispEffect;				// if true, displays the effect
#endif
};


/// Manages all the memory, applications call manager in order to get defined objects from a database.
/// When an object is queried, it first looks in the currently loaded objects. If not there
/// then opens the resource file and gets the object data. 
/// Editor mode:
/// File manager doesn't delete SFObjects and SFRenderObjs when all their reference counts go to 0.
/// The only way these objects can be deleted is through the destroy methods. Textures, on the 
/// other hand, have only one param: their name. The only object that manages textures is the
/// SFMaterial object. A texture can be managed by any material therefore it should never be
/// removed unless all material have dereferenced it. THis cannot be done with two reference counts
/// (regular Release and ReleaseVideoRes) because those are for the video memory and are used
/// by the objects in the game. Therefore textures need a new reference count. This is
/// not needed for SFObjects and SFRenderObj because only one database handles them.
/// Each object's name also needs to be registerd in order to ensure that the name is unique
/// an the file it controls is unique. Whenever an object changes its name it needs to call
/// one of the RegisterX functions, which automatically call that object's FilenameChanged().
/// Materials don't need a reference count because they are either controlled by a render object
/// or by some other object who isn't likely to use the same material for two diffrent objects.
/// Whenver new objects get loaded from the Get methods, they either call LoadMesh or use the
/// data from their file if the bSaved flag is true.
/// Game mode:
/// The game destroys any objects whose reference count is 0. Only SFRenderObj is controlled
/// by the items in the game. The items expect this behavior from SFRenderObj: whenever
/// a call to GetRenderObj() is made, it returns a ready object whose data or system data is all
/// filled out. Then whenever they call LoadVideoRes, the data is loaded into the video card
/// (if not loaded already). A call to Release could remove the object from the system and the
/// object will have to be reloaded again by parsing the database file.

/// Saving and Loading:
/// The game file saves tables with offsets into the file. Since all objects's headers
/// in the editor are loaded, it isn't necessary to ever use these tables.
/// Saves the editor database file as .sfe and the game one as .sfg.
/// The editor database file could be saved over the network if the option is given.
class SFFileManager
{
public:

	/// memory loaded
	typedef map<WORD, SFResource* > MAPSFRESOURCES;
	typedef map<WORD, SFRenderObj* > MAPSFRENDEROBJS;
	typedef map<WORD, DWORD> MAPTABLE;

	SFFileManager();
	~SFFileManager();

	/// in editor, reads all the info into the maps
	void Create(bool bDestroy, TCHAR* pNewDirectory, TCHAR* pNewFilename, TCHAR* pNewGameFilename, DWORD dwNewDB);

	void Destroy(bool bDestroyFilenames);

#ifdef SHADOWFRAME_EDITOR
	void WriteEditor(LPD3DDEV pd3dDevice);
	void ReadEditor();			///< reads all headers and creates all objects

	void Write(LPD3DDEV pd3dDevice);		///< writes tabales
#else
	void Read();							///< reads all tables
#endif

	/// returns a file handle for reading a binary stream that must be closed with fclose()
	FILE* OpenGameFile();

	inline DWORD GetDB()							{ return dwDB; }

#ifdef SHADOWFRAME_EDITOR
	inline TCHAR* GetDirectory()			{ return pDirectory; }
	inline TCHAR* GetDBFilename()			{ return pDBFilename; }
	inline TCHAR* GetGameFilename()			{ return pGameFilename; }

	/// returns a new object, adds a reference count
	SFResource* GetNewResource(bool bSystemMem, WORD wType);

	/// wType is valid when there are different types of SFRenderObj that can be used
	virtual SFRenderObj* GetNewRenderObj(bool bSystemMem, WORD wType);

	/// checks if resource exists, if so returns pointer; if not, creates a resource of wType type; both add a reference and load memory.
	/// NOTE: potential bug when quering for a texture that is already initiated with certain
	/// options like cube map and mip map, but this new one won't be. Explicitly check.
	/// \param wType - one of SFRESOURCE_ flags. If resource with same id exists but not in the specified type, returns NULL
	/// \param wResOpt - extra resource management options like MIP mapping
	/// \param bForceReload - forces the object to reload if an existing one is found
	SFResource* GetResource(const TCHAR* pFilename, LPD3DDEV pd3dDevice, bool bSystemMem, WORD wType = 0, WORD wResOpt = 0, bool bForceReload = false);

	/// changes the object's filenames; first checks if that file is not used by other objects
	/// returns true if successful. Note that this calls ::FilenameChanged, so new resources are loaded
	bool RegisterResource(WORD wId, const TCHAR* pFilename, LPD3DDEV pd3dDevice, bool bSystemMem);

	//@{
	/// destroys from all lists; for textures, checks if the byEditorRef is 0
	void DestroyResource(WORD wId);
	void DestroyRenderObj(WORD wId);
	//@}
	//@{
	inline const MAPSFRESOURCES&		GetResources()		{ return mapSFResources; }
	inline const MAPSFRENDEROBJS&		GetRenderObjs()		{ return mapSFRenderObjs; }
	//@}

	/// returns true if none of the objects have been changed
	bool IsSaved();

	/// Even through their regular reference count might go to 0, resources shouldn't be deleted
	/// unless all the materials release it.
	/// \return true if found the resource and completed the operation
	bool AddEditorRef(WORD wId);
	bool ReleaseEditorRef(WORD wId);

#endif

	//@{
	/// retries, if they exist, the resource objects without trying to load any resources to the video card
	/// \param wRequestType if non zero, only resources of this type will be returned
	SFResource* GetResourcePtr(WORD wId, WORD wRequestType = 0);
	SFRenderObj* GetRenderObjPtr(WORD wId);
	//@}

	//@{
	/// adds a reference to returned object and loads the memory
	/// In game, if wId doesn't exist in lists; looks in the database file and reads
	/// In editor, if can't find in lists, then doesn't exist so return NULL.
	// if wResquestType is one of the SFRESOURCE_ flags (if more than one set, behavior is undefined)
	// if 0 is set, returns the resource without type checking
	// returns NULL if can't find the requested type
	SFResource* GetResource(WORD wId, LPD3DDEV pd3dDevice, bool bSystemMem, WORD wRequestType = 0);
	SFRenderObj* GetRenderObj(WORD wId, LPD3DDEV pd3dDevice, bool bSystemMem);
	//@}

	bool HasResource(WORD wId, WORD wRequestType);
	bool HasRenderObj(WORD wId);

	bool ReleaseResource(SFResource* pResource);
	//bool ReleaseResource(WORD wResId);
	
	// in game, removes from all lists and destroys; in editor, leaves the object
	bool ReleaseRenderObj(SFRenderObj* pRndObj);

protected:

	/// just returns the new pointer, used by GetNewResource and GetNewRenderObj
	virtual SFResource* GetEmptyResource(bool bSystemMem, WORD wType);
	virtual SFRenderObj* GetEmptyRenderObj(bool bSystemMem, WORD wType);

	MAPSFRESOURCES mapSFResources;
	MAPSFRENDEROBJS mapSFRenderObjs;

#ifdef SHADOWFRAME_EDITOR
	WORD wSFResId, wSFRenderObjId;

	/// keeps track of names used so that user doesn't allocated resources with same name 
	/// because they will share the same file
	MAPNAMES mapUsedResNames;

	bool bSaved;
#else
	/// describes the current file
	MAPTABLE mapSFResTable, mapSFRndObjTable;
#endif

	TCHAR* pDirectory;							///< directory where the database is stored in
	TCHAR *pGameFilename;	
	TCHAR* pDBFilename;							///< editor database file
	DWORD dwDB;
};

/// rebuilds vecTexIds and vecTexRes in order to contain the all the textures in vecNewConsts
/// vecTexIds and vecTexRes can contain old unreleased texture info, which will be released. Before releasing
/// make sure to load the new textures first (so that unnecessary memory loading is avoided).
/// The SHADERCONST::pd3dtex member in vecNewConsts will also be released before reloaded
HRESULT ExtractTextures(LPD3DDEV pd3dDevice, SFFileManager* pMngr, vector<SHADERCONST*>& vecNewConsts, bool bRef, bool bLoadRef);

#endif