#ifndef SFFILEMANAGER_OBJECT
#define SFFILEMANAGER_OBJECT

#define OBJOPT_NORMAL			0x0002		///< recompute normals of new mesh
#define OBJOPT_TANGENT			0x0004		///< has tangent info
#define OBJOPT_SOFTSHADOW		0x0008		///< has soft shadows

/// offsets from start of vertex data, pos is always at 0 and is always present
struct DATAOFFSET
{
	BYTE byBlend;
	BYTE byIndex;
	BYTE byNormal;
	BYTE byTex;
};

struct COMPRESSIONINFO
{
	COMPRESSIONINFO() { pmat = NULL; }
	~COMPRESSIONINFO() { delete pmat; }

	DXMAT* pmat;			///< if( dwData & CTPOS ) transformation from compression space to normal space
};

DATAOFFSET GetDataOffset(DWORD dwData);
DWORD GetDataSize(DWORD dwData);
DWORD GetDataFromFVF(DWORD dwFVF);
DWORD GetFVFFromData(DWORD dwData);		///< ignores compression; if has tangent, uses
										///< texture stage 0 for it and shifts all the other textures

D3DVERTEXELEMENT9* GetVertexShaderDecl(DWORD dwData, DATAOFFSET* pdoff = NULL);

#ifdef SHADOWFRAME_EDITOR
void CompressVertexBufferFmt(BYTE* pSrc, BYTE*& pDst, LPD3DDEV pd3dDevice, WORD wNumVerts, 
							 DWORD dwSrcData, DWORD dwDstData, DWORD dwExtraBoneIndices, COMPRESSIONINFO* pci);

// creates an edge list of positions only, also compresses the edge list
// for the sake of not adding another parameter, the stride of the new vb is not returned
HRESULT CreateSoftShadow(BYTE* pSrc, WORD* pIndices, LPD3DDEV pd3dDevice, WORD wNumVerts, DWORD dwVertData, 
						 COMPRESSIONINFO* pciVert, DWORD dwNumFaces, WORD& wNumShadowVerts, COMPRESSIONINFO* pci,
						 float fNormExtend, float fBaseExtend, float fEdgeExtend, float fSelfShadow,
						 LPD3DVB* pvbEdges, BOOL bSaveWeights);

// insiparation for this fn: Animations use vertex data which means that they must load
// a mesh and compress it. THis vertex data cannot be a separate SFObject because
// that is totally unnecessary and will slow things down (hmm...)
// the vertices are computed by: fSize * v - vOffset
/// dwData is filled with the foramt of the mesh, it already contains the compression info for compressing
HRESULT LoadStandardMesh(const TCHAR* pFilename, D3DPOOL d3dpool, DWORD usg, LPD3DDEV pd3dDevice, COMPRESSIONINFO* pci,
						 LPD3DVB* ppvb, LPD3DIB* ppib, DWORD& dwData, BYTE& byFVFSize, AABB* pab, float* pfRadius,
						 WORD& wNumVerts, DWORD& dwNumFaces, vector<D3DXATTRIBUTERANGE>* pvecAttribs, float fSize, DXVEC3& vOffset,
						 BOOL bSaveWeights, BOOL bTangent, BOOL bNormalRecompute);
#endif

HRESULT CreateIndexedQuadIB(LPD3DDEV pd3dDevice, WORD wNumQuads, LPD3DIB* pib);

/// Holds memory to a simple object and its shadows (if it supports them)
/// Manages its own file to save its resource data and options. Whenever building
/// new shaders, call WriteShaderData() in order to decompress and animate the positions
/// and write the input structure semantics.
class SFObject : public SFResource
{
	friend class SFFileManager;
	friend class SFSpaceFileManager;
	friend class SFRenderObj;
	friend class MotherShipClassData;

	friend class CSFObjectDlg;			///< editor only

public:
	SFObject(WORD wResType);
	virtual ~SFObject();

	virtual LPD3DVB GetVB(int attrib) = 0;
	virtual LPD3DIB GetIB(int attrib) = 0;

	virtual LPD3DVB GetSystemVB(int attrib) = 0;
	virtual LPD3DIB GetSystemIB(int attrib) = 0;

	virtual D3DXATTRIBUTERANGE* GetAttrib(int attrib) = 0;
	virtual int GetNumAttribs() = 0;

	inline bool IsTangent()					{ return (wResOpt&OBJOPT_TANGENT)>0; }
	inline bool IsSoftShadow()				{ return (wResOpt&OBJOPT_SOFTSHADOW)>0; }

#ifdef SHADOWFRAME_EDITOR
	virtual HRESULT SetCompression(DWORD dwNewCmp, TCHAR* pCurDirectory, LPD3DDEV pd3dDevice) = 0;
	
	/// only processes shadow if the main vertex buffer is allocated
	virtual HRESULT SetSoftShadow(bool bNewSoftSh, LPD3DDEV pd3dDevice) = 0;
	virtual HRESULT SetGenerateTangent(bool bNewTangent, TCHAR* pCurDirectory, LPD3DDEV pd3dDevice) = 0;
#endif

	inline ColObj& GetColObj()				{ return cobj; }

	inline AABB& GetAABB()					{ return ab; }
	inline float GetRadius()				{ return fRadius; }
	inline DWORD	GetData()				{ return dwData; }

	virtual DWORD GetNumFaces() = 0;
	virtual WORD GetNumVerts() = 0;

	inline BYTE		GetFVFSize()			{ return byFVFSize; }
	inline LPD3DDECL	GetDecl()			{ return pd3dDecl; }

#ifdef SHADOWFRAME_EDITOR
	ColModelObj&		GetPreciseCol()		{ return cobjPrecise; }
#endif

protected:
	AABB ab;				///< bounds the cobj, for animations this bounds all animations
	float fRadius;			///< radius of object centered at ab.vPos

	ColObj cobj;			///< bounding boxes for model
							///< animation won't be using ColObj because animation changes all the time and
							///< there won't be one consistent cobj
							///< SFAnimControl contains the collision data for animations

#ifdef SHADOWFRAME_EDITOR
	ColModelObj cobjPrecise;		///< precise collision info based on all faces, used to move the object with the mouse
#endif

	LPD3DDECL pd3dDecl;							///< the declaration (initiated and released on Load(Release)VideoRes calls)
	DWORD dwData;						///< low 16 bits describe the input format, high 16 bits describe compression used

	BYTE byFVFSize;						///< uncompressed and compressed vertex sizes of the buffers
	COMPRESSIONINFO ci;				///< compression info like transformation matrices etc

#ifdef SHADOWFRAME_EDITOR
	/// initial offset and scales
	DXVEC3 vOffset;
	float fScale;						///< used for size of object at the moment
#endif
};

class SFMeshObject : public SFObject
{
	friend class SFFileManager;
	friend class SFSpaceFileManager;
	friend class SFRenderObj;
	friend class MotherShipClassData;

protected:

	struct OBJECTMEM
	{
		OBJECTMEM();
		~OBJECTMEM();

		LPD3DVB pvb, pvbShadow;
		LPD3DIB pib;
	};

	SFMeshObject();
	virtual ~SFMeshObject();

	virtual void DestroyRes();

public:

	virtual HRESULT LoadVideoRes(LPD3DDEV pd3dDevice);
	virtual void ReleaseVideoRes();

#ifdef SHADOWFRAME_EDITOR
	//@{
	/// Functions defined only by the editor.

	virtual TCHAR* GetSaveExt();
	virtual HRESULT SetCompression(DWORD dwNewCmp, TCHAR* pCurDirectory, LPD3DDEV pd3dDevice);
	
	/// only processes shadow if the main vertex buffer is allocated
	virtual HRESULT SetSoftShadow(bool bNewSoftSh, LPD3DDEV pd3dDevice);
	virtual HRESULT SetGenerateTangent(bool bNewTangent, TCHAR* pCurDirectory, LPD3DDEV pd3dDevice);

protected:
	/// releases all memory and loads a mesh from pFilename param
	/// if bSoftShadow is true, loads the soft shadow
	virtual HRESULT Load(LPD3DDEV pd3dDevice, bool bSystemMem);

	virtual void WriteHeader(FILE* f);

	virtual void WriteHeaderEditor(FILE* f);
	virtual void ReadHeaderEditor(FILE* f);

	virtual bool WriteResData(LPD3DDEV pd3dDevice);

public:
	//@}
#endif

	virtual void ReadHeader(FILE* f);

#ifdef SHADOWFRAME_EDITOR
protected:
#endif
	/// Reads the binary file from file and loads into system mem if bSystemMem is true
	/// or into video mem if false.
	virtual HRESULT ReadResDataFile(FILE* f, LPD3DDEV pd3dDevice, bool bSystemMem);

public:
	//@{
	/// Shader specific fns.

	/// Fills the input structure and loads the obj data into the appropriate variables.
	/// uses the SHADERINPUT_ strings for the variables
	virtual void WriteShaderData(sfstring& sfinput, sfstring& sfvscode);
	virtual void SetShaderConst(ISFEffect* peff);		// set compression matrix if used
	//@}

	//@{
	/// Get various obj data.

	/// transformation of input data to real object data; if NULL, no transformation
	inline DXMAT*	GetTrans()				{ return ((dwData&CTPOS) && ci.pmat) != NULL ? ci.pmat : NULL; }
	virtual LPD3DVB GetSoftVB(size_t attrib)  { return mem.pvbShadow; }

	virtual DWORD	GetNumFaces()			{ return dwNumFaces; }
	virtual WORD		GetNumVerts()			{ return wNumVerts; }

	virtual LPD3DVB GetVB(int)					{ return mem.pvb; }
	virtual LPD3DIB GetIB(int)					{ return mem.pib; }

	virtual LPD3DVB GetSystemVB(int)			{ return pSystemMem != NULL ? pSystemMem->pvb : NULL; }
	virtual LPD3DIB GetSystemIB(int)			{ return pSystemMem != NULL ? pSystemMem->pib : NULL; }

	virtual D3DXATTRIBUTERANGE* GetAttrib(int attrib)	{ return &vecAttribs[attrib]; }
	virtual int GetNumAttribs() { return vecAttribs.size(); }
	//@}

protected:
	vector<D3DXATTRIBUTERANGE> vecAttribs;		// the attributes of the object, if faces divided
												// needs to be saved in the header since the number of possible materials a
												// render object can implement depends on its size
	OBJECTMEM mem;
	OBJECTMEM* pSystemMem;

	DWORD dwNumFaces;
	WORD wNumVerts, wNumShadowVerts;
	
	BYTE byShadowFVFSize;				///< size of the shadow vertex size
};

#endif
