#ifndef SF_RENDER
#define SF_RENDER

// extacts the handle of the const and stores it in mapHandles
// append a 't_' before the name for textures, needs a 40 byte char array called mbstr
#define EXTRACT_SHADERHANDLE(id, pEffect) \
{ \
	if( pEffect != NULL ) { \
		if( pconst->UsesTex() ) { \
			strncpy(mbstr, "t_", 2); \
			StringCchCopyA(mbstr+2, 38, pconst->GetMBName()); \
			pEffect->SetHandle(id, mbstr); \
		} \
		else pEffect->SetHandle(id, pconst->GetMBName()); \
	} \
} \

/**
 The render manager handles the construction of the effects files necessary to render object.
 Every light and object in the scene must be registered with the manager in order to get handled
 properly.
 In order to construct an effect file for an object, the manager handles shader fragments that define
 a particular variable. Whenever this variable is used by the object's material, the code to define
 that variable will be inserted in the effect file. The manager also handles the constant memory.
 Different objects like SFObject and SFLight can define their own constants. To manage this, 
 every header file needs to define a GetRegisteredConsts function that
 inserts the necessary constants. Every time the effect file is built, it will need to extract handles 
 to all the constants that it defines. This will be handled as a map. The ids of the constants
 are assigned statically, and the manager makes sure that no ids overlap
*/

// global constants
#define SHC_WORLDVIEW		(SHCONST_USER+1)
#define SHC_WVP				(SHCONST_USER+2)
#define SHC_ELAPSEDTIME		(SHCONST_USER+3)
#define SHC_CAMERAPOS		(SHCONST_USER+4)
#define SHC_CAMERADIR		(SHCONST_USER+5)
#define SHC_SUNDIR			(SHCONST_USER+6)
#define SHC_OBJECTMAT		(SHCONST_USER+11)
#define SHC_INVWORLDMAT		(SHCONST_USER+12)

// different wind directions
#define SHC_WIND1			(SHCONST_USER+7)
#define SHC_WIND2			(SHCONST_USER+8)
#define SHC_POINTWIND1		(SHCONST_USER+9)

#define WORLDVIEW_NAME		_T("c_WorldView")
#define VIEWPROJ_NAME		_T("c_ViewProj")
#define WVP_NAME			_T("c_WVP")			///< world view projection matrix
#define ELAPSEDTIME_NAME	_T("c_ElapsedTime")
#define CAMERAPOS_NAME		_T("c_CameraPos")
#define CAMERADIR_NAME		_T("c_CameraDir")
#define SUNDIR_NAME			_T("c_SunDir")
#define WORLDMAT_NAME			_T("c_WorldMat")
#define INVWORLDMAT_NAME		_T("c_InvWorldMat")

#define WIND1_NAME			_T("c_Wind1")
#define WIND2_NAME			_T("c_Wind2")
#define POINTWIND1_NAME		_T("c_PointWind1")

class SFRenderMngr
{
public:
	SHADERCONST* pcWorldView, *pcWVP, *pcWorldMat, *pcInvWorldMat, *pcElapsedTime, *pcCameraPos, *pcCameraDir, *pcSunDir, *pcWind1, *pcWind2, *pcPointWind1,
		*pcViewProj;

	typedef map<WORD, SFFRAGMENT*> MAPFRAGMENTS;
	typedef map<WORD, SFMaterialType*> MAPMATERIALS;
	typedef map<WORD, SHADERCONST*> MAPCONSTS;
	typedef map<DWORD, SFWorldObj*> MAPWORLDOBJS;
	typedef list<SFLight* > LIGHTLIST;

	/// Describes an effect file to build
	struct SFBUILDEFFECT
	{
		PSINPUTLIST listPSInputs;	///< describes the input names and their dimensions that need to be passed
									///< to the pixel shader from the vertex shader. 
									///< the second member in each pair is defined by FRAG_ flags
		set<WORD> setUsedFrags[2];	///< used fragments in shader [0] - vs; [1] - ps
		set<WORD> setUsedConsts;	///< global constants only

		typedef vector<SFLight::LIGHTSHADERID> LIGHTS;

		SFBUILDEFFECT() : wMtrlId(0) {}

		sfstring streffect;			///< the whole effect
		sfstring strcode[2];		///< Since will be writing simultaneously to global, vs, and ps code, need to have
									///< 3 separate strings. SFBUILDEFFECT::streffect will hold the global code
									///< while strcode[0] is the vs code and strcode[1] is the ps code

		WORD wMtrlId;				///< the material id
		LIGHTS vecLights;
	};

	class SFRenderTarget
	{
	public:
		SFRenderTarget();
		virtual ~SFRenderTarget();

		virtual HRESULT OnResetDevice(LPD3DDEV pd3dDevice);		// inits targets and vb
		virtual void Destroy();

		short index;				// mrt index
		bool bcleared;
		WORD width, height;
		D3DFORMAT fmt;
		LPD3DSURF psurf;
		LPD3DTEX ptex;
		LPD3DVB pvbquad;			// quad that surrounds the whole render target
	};

	// target for post processing, will contain two render targets that will be flipped
	// every time a render operation is called. psurfnext will be the texture while
	// psurf will always be the render surface (change roles when Flip is called)
	class SFPostTarget : public SFRenderTarget
	{
	public:
		SFPostTarget();
		~SFPostTarget();

		HRESULT OnResetDevice(LPD3DDEV pd3dDevice);
		void Destroy();

		// return the texture that can be used to access the surface
		inline LPD3DTEX GetTex() { return bFeedback ? ptexnext : ptex; }

		void Flip();		// flips the current and previous targets if feedback is enabled

		bool bFeedback;	
		LPD3DSURF psurfnext;
		LPD3DTEX ptexnext;
	};

	/// the compiled effect file, has handles for all the variables defined
	/// The effect is unique based on mat id, lights, and environmental effects.
	/// TODO: Should not be unique with respect to compression and obj bone animation because
	/// too many shaders will get created.
	/// Parameters like time are shared across effects. Each effect implements
	/// a main technique tech0 that will be the highest supported technique
	class SFEffect : public ISFEffect
	{
		friend class SFRenderMngr;
	public:

		SFEffect()				{ pstrEffect = NULL; peff = NULL; }
		~SFEffect()				{ SAFE_RELEASE(peff); SAFE_DELETE(pstrEffect); }

		WORD GetId() { return wId; }

		D3DXHANDLE Get(const WORD wId) {
			map<WORD, D3DXHANDLE>::iterator it = mapHandles.find(wId);
			if( it == mapHandles.end() ) return NULL;
			return it->second;
		}

		bool Set(const WORD wId, void* pdata, DWORD dwSize) {
			map<WORD, D3DXHANDLE>::iterator it = mapHandles.find(wId);
			if( it == mapHandles.end() ) return false;
			peff->SetValue(it->second, pdata, dwSize);
			return true;
		}

		void SetHandle(WORD wId, const char* pname) {
			if( peff == NULL ) return;
			D3DXHANDLE h = peff->GetParameterByName(NULL, pname);
			mapHandles[wId] = h;
		}

		WORD GetMtrlId() const				{ return wMtrlId; }
		ID3DXEffect* d3dxEff() const		{ return peff; }
		const TCHAR* GetCode() const		{ return pstrEffect; }

		bool UsesLight(DWORD dwLightId) {
			for(SFBUILDEFFECT::LIGHTS::iterator it = vecLights.begin(); it != vecLights.end(); ++it) {
				assert( it->plight != NULL );
				if( it->plight->GetId() == dwLightId ) return true;
			}
			return false;
		}

		HRESULT OnResetDevice() { if( peff != NULL ) return peff->OnResetDevice(); return S_OK; }
		HRESULT OnLostDevice() { if( peff != NULL ) return peff->OnLostDevice(); return S_OK; }

		inline const vector<WORD>& GetEffectChans() { return vecEffectChans; }

	private:
		//@{
		/// Saves the binary compiled shader files for every technique, then
		/// when read, sets the shaders to the effect.
		void Write(FILE* f);
		void Read(FILE* f);
		//@}

		DWORD dwData;			///< SFObject::dwData type of format for obj being rendered (\see sffilemanager_object.h)
		SFBUILDEFFECT::LIGHTS vecLights;

		WORD wMtrlId;						///< mat id coding for
		vector<WORD> vecEffectChans;		///< effect channels rendering to for every pass

		TCHAR* pstrEffect;		///< the text file that defines the ps and vs fns with all the techniques
		ID3DXEffect* peff;

		map<WORD, D3DXHANDLE> mapHandles;		///< all handles of all constants that are defined
		WORD wId;								///< unique id of the effect
	};

	// A post process effect will write to its own specified render target and use other render targets. This is
	// controlled through annotations on the technique and texture variables declared.
	// If width, height are specified in the technique then the dimensions of the new target will match those
	// If scalex and scaley are specified, then the dims of the new target will be that scale value of the current render targ (take precedence)
	// postrt specifies the render targ id (default is -1)
	// for variables:
	// rt and postrt: indicate which render target id to attach to that texture.
	// ConvertPixelsToTexels: converts pixel offsets from SourceVariable to texel offsets, format: ConvertPixelsToTexels = "SourceVariable";
	class SFPostProcessEffect : public ISFEffect
	{
		friend class SFRenderMngr;
	public:

		SFPostProcessEffect()				{ peff = NULL; index = -1; }
		~SFPostProcessEffect()				{ SAFE_RELEASE(peff); }

		WORD GetId() { return wId; }

		D3DXHANDLE Get(const WORD wId) {
			map<WORD, D3DXHANDLE>::iterator it = mapHandles.find(wId);
			if( it == mapHandles.end() ) return NULL;
			return it->second;
		}

		bool Set(const WORD wId, void* pdata, DWORD dwSize) {
			map<WORD, D3DXHANDLE>::iterator it = mapHandles.find(wId);
			if( it == mapHandles.end() ) return false;
			peff->SetValue(it->second, pdata, dwSize);
			return true;
		}

		void SetHandle(WORD wId, const char* pname) {
			if( peff == NULL ) return;
			mapHandles[wId] = peff->GetParameterByName(NULL, pname);
		}

		WORD GetMtrlId() const				{ return 0; }
		ID3DXEffect* d3dxEff() const		{ return peff; }
		const TCHAR* GetCode() const		{ return strfilename.c_str(); }
		bool UsesLight(DWORD dwLightId)		{ return false; }

		HRESULT OnResetDevice();		// have to conver all the values with ConvertPixelsToTexels annotation
		HRESULT OnLostDevice() { if( peff != NULL ) return peff->OnLostDevice(); return S_OK; }

		short index;					///< render target index to vecPostRTs, if -1 effect is special
	private:
		sfstring strfilename;
		ID3DXEffect* peff;

		list<string> listPPEffectsUsing;					///< name of effects files that have to be rendered before this (currently only for scene.fx)
		list<WORD> listPPEffectsIds;						///< corresponding ids of listPPEffectsUsing

		list< pair<D3DXHANDLE, WORD> > listRTUsing;		///< if the high bit is 1, this is a post process render targ, otherwise regular targ
		list< pair<D3DXHANDLE, D3DXHANDLE> > listTexelPairs;	///< the first pair is the converting variable while the second is 
																///< the source dx,dy vals in pixel coordinates of dest surface
		list< pair<D3DXHANDLE, D3DXHANDLE> > listSourcePairs;	///< same as listTexelPairs, except everything is specified in the original render target pixels

		bool bRendered;					///< true if rendered this frame
		WORD wReqWidth, wReqHeight;		///< the requested width and height

		WORD wId;				///< unique id of the effect

		map<WORD, D3DXHANDLE> mapHandles;		///< all handles of all constants that are defined
	};

	SFRenderMngr();
	~SFRenderMngr();

	inline LPD3DDEV GetDevice()				{ return pd3dDevice; }
	static SFRenderMngr& GetMngr()			{ assert( pRenderMngr != NULL); return *pRenderMngr; }

	void Destroy();								///< destroys device and releases all video memory memory
												///< purpose: when exiting the game
	void Reset();								///< destroys all object/lights their shader sequences and any built shaders
												///< purpose: used when switchign between modules

	/// loads all effect files and inits the core engine elements
	HRESULT InitDeviceObjs(LPD3DDEV g_pd3dDevice, const TCHAR* pfilename, WORD wRendWidth, WORD wRendHeight); 

	HRESULT OnResetDevice();		///< loads device dependent data
	void OnLostDevice();			///< releases the device dependent data
	void DeleteDeviceObjs();		///< destroys all data

	EDITOR(bool SaveEngine(const TCHAR* pfilename);)				///< saves the materials/fragments/lights and any other data 

	void AddLight(SFLight*);
	void RemoveLight(SFLight*);

	void AddObj(SFWorldObj*);
	void RemoveObj(SFWorldObj*);
	bool ContainsObj(DWORD dwObjId)		{ return mapWorldObjs.find(dwObjId) != mapWorldObjs.end(); }

	HRESULT RebuildTextures();

	// creates a post process effect and make sure that the render target it programs for is valid, if an effect with the same name
	// exists, delete it
	HRESULT CreatePostProcessEffect(const TCHAR* pfilename, SFPostProcessEffect** peff = NULL, string* pstrDebug = NULL);

	// the global constants are used by the post process effects, changing them means that the constants in the effects have to be changed
	void OnChangePostProcessConsts();

	SFPostProcessEffect* GetPostProc(const TCHAR* pfilename);
	inline SFPostProcessEffect* GetPostProc(WORD wId) { if( mapPostEffects.find(wId) != mapPostEffects.end() ) return mapPostEffects[wId]; return NULL; }

	/// creates all the effects that an object will use
	HRESULT CreateEffect(SFWorldObj* pobj);

	/// Given the material and the input declaration will create the appropriate effect
	/// \param material id
	/// \param pInputDecl string describing the vertex shader input declaration struct (ie. float4 v_pos : POSITION; )
	///					The standard input names are defined in sffilemanager.h
	/// \param pVSDecl string describing the code put in the beginning of the vertex shader to init the beginning fragments (names defined in sffilemanager.h)
	/// \param vecLights the lights that should be inserted in this effect
	/// \param bUsePP is true, try rendering with per pixel effects first (if material supports them)
	HRESULT CreateEffect(WORD wMtrlId, const TCHAR* pInputDecl, const TCHAR* pVSDecl, vector<SFLight::LIGHTSHADERID>& vecLights, bool bUsePP, ISFEffect** ppNewEff);


	inline void SetChannels(DWORD dwChanMask);			// from SFEffect::vecEffectChans
	/// Should be called inside device->Begin/End calls @{

	/// resets buffers and prepares for new scene
	/// \param {mView, mProj, and frView} the camera that will be used for the rendering
	HRESULT BeginScene(const DXMAT& mView, const DXMAT& mProj, const FRUSTUM& frView, float fTimeElapsed, DWORD dwClearColor = 0);

	/// if bLoadVB is false, the vertex and index buffers are assumed to be set, if true and resources are not loaded, will load the resources of the obj
	HRESULT Render(SFWorldObj* p, bool bLoadVB = true, RenderPass passtype = RP_Normal);
	HRESULT BeginEffect(SFEffect* peff, DXMAT& mtrans, DXMAT& minvtrans);		///< sets up the effect and load all constants. After this, setup any extra constants and call
																///< ID3DXEffect::Begin, BeginPass, EndPass and any DrawPrimitive method

	/// Reflections
	/// All reflections for the scene will render onto one surface: ptargReflection
	/// This might cause problems if two different reflection surfaces are overlapping

	/// Clears stencil buffer and sets up the correct render states to render the reference object, returns the stencil ids that will be written
	// disables writes since if enabled, the reflected geometry won't be drawn. This can be combined with the render to zbuffer pass
	WORD BeginReflectionStencil();
	void EndReflectionStencil();

	/// After stencil buffer is filled, call in order to render only on stencil regions and set appropriate render target
	HRESULT BeginReflections();

	/// All objects in front of vplane are reflected, and all objects behind it are ignored. Pushes the previous proj matrix in listPrevView
	/// Called once for every new reflection surface
	/// \param wStencilId returned from BeginReflectionStencil
	HRESULT RenderReflection(DXVEC4& vplane, WORD wStencilId);


	// restores all previous states and render targets
	HRESULT EndReflections();

	/// renders the depth map only (disables all writes to color) into plight->GetDepth()
	HRESULT RenderShadowMap(SFLight* plight);

	/// Gathers all render targets and does post processing finishing with ppSceneEffect, since this will combine all post processing effects,
	/// all other non material geometry like particle effects need to be rendered into its primary buffer beforehand.
	/// if iSceneRendTarg is != -1, then will render to vecPostRTs[iSceneRendTarg] instead of final original target. In the end, restores
	/// the original render target.
	HRESULT EndScene(short iSceneRendTarg = -1);	
	//@}

	/// called to render a post processing effect
	/// \param bUseQuad if true, will render a quad that covers the whole render target
	///					if false, will only set the constants and leave the rendering to be done outside
	///					if so, remember to flip the target with FlipPPTarg
	HRESULT RenderPost(SFPostProcessEffect* ppost, bool bUseQuad = true);

	void FlipPPTarg(WORD wId)				{ if( wId < vecPostRTs.size() && vecPostRTs[wId] != NULL ) vecPostRTs[wId]->Flip(); }

	// render target querying
	LPD3DSURF GetChannelSurf(WORD windex)		{ return  (windex < vecEffectChans.size() && vecEffectChans[windex] != NULL) ? vecEffectChans[windex]->psurf : NULL; }
	void SetWorldMat(D3DXMATRIX& m, D3DXMATRIX& inv, SFEffect* pSetInEffect = NULL);
	// provide different matrices
	void SetWorldMat(D3DXMATRIX& m, D3DXMATRIX& inv, D3DXMATRIX& mTempProj, SFEffect* pSetInEffect = NULL);

	// render primitives like fragments, constants, and materials and their create/update fns
	LPD3DIB GetQuadIndices(DWORD dwNumQuads);
	inline DWORD GetNumQuadsInIB() { return dwNumIndexedQuads; }

	inline SFFRAGMENT* GetFragment(WORD wId) {
		MAPFRAGMENTS::iterator it = mapFragments.find(wId);
		if( it == mapFragments.end() ) return NULL;
		return it->second;
	}

	SHADERCONST* GetGlobalConst(WORD wId)			{
		if( wId == 0 ) return NULL;
		MAPCONSTS::iterator it = mapShaderConsts.find(wId);
		return it != mapShaderConsts.end() ? it->second : NULL;
	}

	SHADERCONST* GetGlobalConst(const TCHAR* pConstName);

	/**
		All add/update/remove functions recheck the validity of the rest of the objects in editor mode (game mode assumes no such dependencies). This can
		be slow, so have two modes of the render manager: interactive and developer. In developer mode,
		the new changes won't be propagated to other objects and there will be no rendering. In interactive
		mode, all changes will be propagated across the manager and objects will be rendered.
	*/
	SFMaterialType* CreateMaterial(WORD wTemplateMtrl = 0);		///< if wTemplateMtrl is a valid mtrl id, will init the new material to have the
																///< same consts and code as the template material
	inline SFMaterialType* GetMtrl(WORD wId) {
		MAPMATERIALS::iterator it = mapMtrls.find(wId);
		if( it == mapMtrls.end() ) return NULL;
		return it->second;
	}

	SFFRAGMENT* AddGlobalFragment(const TCHAR* pname);

	/// searches for the fragments name, and changes the stored pointer, makes sure that it has a unique id
	/// This is also used as an edit function for fragments since it will change all objs using the replaced fragment
	/// \param bInheritId if true and a fragment with the same name is found, will use the id of the old fragment
	/// \return true if fragment successfully added
	bool AddFragment(SFFRAGMENT* pfrag, bool bIheritId = false);

	inline SFFRAGMENT* GetFragment(const TCHAR* pname)
	{
		MAPFRAGMENTS::iterator it = mapFragments.begin();
		while(it != mapFragments.end()) {
			if( it->second->strname.c_str() != NULL && _tcscmp(it->second->strname.c_str(), pname) == 0 ) return it->second;
			++it;
		}
		return NULL;
	}

	/// Add only global constants and not per material/per light type consts
	SHADERCONST* AddGlobalConst(const TCHAR* pConstName);	///< adds/edits a constant and inits its id
	bool AddConst(SHADERCONST* pconst);				///< adds a constant with id already initiated (make sure that id along with name are unique)
													///< if constant exists with same id, replace it
													///< Uses the same pointer as the one passed in, so don't delete

#ifdef SHADOWFRAME_EDITOR
	void RemoveMaterial(WORD wMtrlId);
	void RemoveFragment(WORD wId);
	bool RemoveGlobalConst(WORD wId);

	void Revalidate(bool bLoad);			// rebuilds all the valid flags and shader dependencies
	void CheckValidity(SFFRAGMENT* pfrag);
	void RevalidateFragments(bool bAll);				///< if bAll is false validates all those fragments whose wValid value is 0

	/// Rebuilds dependencies, and fills any valid flags. Resets all objs that are using this material and updates their constants.
	/// Should be called when a material constant is added or its code is changed. Necessary to call when the constant types change too.
	/// Every SFRenderObj is updated with the new material
	/// \param bUpdateObjConsts if true, will go through every SFRenderObj and update its constants
	HRESULT OnMaterialChanged(SFMaterialType&, bool bLoad, bool bUpdateObjConsts = true);

	///< sets to changed all the materials and light types that are using this variable 
	void OnVarChanged(const TCHAR* pname);
	void OnLightTypeChanged(WORD wType);	///< validates the new type (preprocesses the code) and rebuilds all lights that are of this type
	void OnLightTypeRemoved(WORD wType);	///< reset all lights using this type
#endif

	inline WORD GetRendWidth()		{ return wRendWidth; }
	inline WORD	GetRendHeight()		{ return wRendHeight; }
	inline LPD3DSURF GetOrgRendTarg() { return pOrgRenderTarg; }
	inline LPD3DSURF GetOrgZSurf() { return pOrgZSurf; }

	inline const MAPCONSTS& GetConsts() { return mapShaderConsts; }
	inline const MAPFRAGMENTS& GetFragments() { return mapFragments; }
	inline const MAPMATERIALS& GetMaterials() { return mapMtrls; }

	inline const MAPWORLDOBJS& GetWorldObjs() { return mapWorldObjs; }

	inline D3DXMATRIX& GetView() { return mView; }
	inline D3DXMATRIX& GetProj() { return mProj; }
	inline FRUSTUM& GetFrView() { return frView; }
	inline void SetFrView(FRUSTUM& fr) { frView = fr; frView.GetMatrices(mView, mProj); }

	HRESULT SetEffectChannel(DWORD ichan);		// sets the render target to the specified effect channel
	LPD3DTEX GetReflTex() { return ptargReflection != NULL ? ptargReflection->ptex : NULL; }

	ID3DXEffect* GetSimpleEffect() { return peffSimpleShader; }

	inline void SetAnimSun(bool banim) { bAnimSunDir = banim; }
	inline bool GetAnimSun() { return bAnimSunDir; }

private:

	/// Write shader code to a vertex or a pixel shader. The code is already assumed valid
	/// for writing to the selected shader. In order for a write to succeed, the
	/// fragments the shader code depends on must be valid. 
	///		- bPS, then each depending fragment must have FRAG_COMPPS or FRAG_PASSTOPS flag
	///		- !bPS, then each fragment must have FRAG_COMPVS flag
	/// \param bPS true if the code should be written in the pixel shader
	/// \param pLightId optional id to attach at the end of variables sent to the ps with SendToPSLight
	/// \return true if all code written. False will occur when the fragments the code
	/// depends on are not valid
	bool WriteCodeToEffect(SFBUILDEFFECT& build, SHADERCODE& code, bool bPS, TCHAR* pLightId = NULL);

	/// write code for a light. Don't call SFLight::AddConstIndices to add the const indices dependent on the light
	/// Also write the listLightOutputs to the vs output semantics and add in the declaration code in the ps
	/// \param wlightid		the id of the light so can rename constants appropriately
	///						if 0, this is the first light that is processed in the shader. All the declarations and fragments 
	///						that all lights will use will be written
	bool WriteCodeToLight(SFBUILDEFFECT& build, SFLight* plight, WORD wlightid, SHADERCODE& code, bool bPS);

	/// Writes all the fragments that haven't been written yet by calling WriteCodeToEffect
	/// \param build Holds the already written fragments
	/// \param setNewFrags The new fragments
	/// \param bPS whether the fragments should be written in ps or in vs
	bool WriteFragments(SFBUILDEFFECT& build, set<WORD>& setNewFrags, bool bPS, TCHAR* pLightId = NULL);

	/// fills a list of the fragments that the code depends on.
	/// \param plistUnknown if not null fills it with the names of the variables that it did not recognize (pointers to pcode)
	/// \return true if there are unknown variables
	bool FillDependencies(const TCHAR* pcode, set<WORD>& setdep, list<const TCHAR*>* plistUnknown);

	/// Adds the global constants used from mapShaderConsts. All the unknown constants are returned in plistUnknown. In order to
	/// avoid getting unknown constants that belong to the materials or other sources like light, enter the material or another extra vector.
	/// The constants found to belong to the optional sources will not be added to setdep.
	/// \param pmtrl optional material that defines more constants
	/// \param pvecConsts optional vector of SHADERCONST* that defines more constants
	/// \return true if there are unknown constants
	bool FillConstantsUsed(const TCHAR* pcode, SFMaterialType* pmtrl, SHADERCONSTVEC* pvecConsts, set<WORD>& setdep, list<const TCHAR*>* plistUnknown);

	/// used for asserting
	bool LightInList(SFLight* pHikari);

	LIGHTLIST listHikari;						///< all the lights in the scene
	MAPWORLDOBJS mapWorldObjs;					///< the current objects that are going to be rendered

	LPD3DIB pibSoftShadow;						///< indices for the soft shadow
	DWORD dwNumIndexedQuads;					///< num faces that pibSoftShadow supports
	
	WORD wNextConstId, wNextFragId, wNextMtrlId;
	DWORD dwNextEffectId;
	FRUSTUM frView;								///< frustum of the camera, modified by the cam variables in ::Move
	DXMAT mView, mProj;

	list<DXMAT> listPrevView;					///< previous projections with the tail matrix being the most recent,
												///< used in reflections
	list<FRUSTUM> listPrevFrustum;

	/// render device variables
	LPD3DDEV pd3dDevice;
	WORD wRendWidth, wRendHeight;				///< extents of the render surface

	WORD wShadowWidth, wShadowHeight;		// for now const shadow dims
	LPD3DSURF psurfShadow;					// the render target when rendering a shadow
	D3DFORMAT fmtShadow;					///< depth buffer format for shadow	

	/// 3d engine database
	MAPCONSTS mapShaderConsts;					///< registered constants
	MAPFRAGMENTS mapFragments;
	MAPMATERIALS mapMtrls;

	vector<SFRenderTarget*> vecEffectChans;		///< render targets that the materials will render to
	vector<SFPostTarget*> vecPostRTs;			///< general render targets to be used the by the post processing effects (they will be init the the requirements on the post processing effetcs)
	SFRenderTarget* ptargReflection;			///< target used for holding reflections
	WORD wPrevStencilVal;						///< previous stencil reference value used for reflection

	ID3DXEffect* peffShared;					///< shared consts effect
	map<DWORD, D3DXHANDLE> mapSharedHandles;

	ID3DXEffect* peffSimpleShader;				///< simple shader that just transforms and outputs a position (uses c_WVP)

	map<WORD, SFPostProcessEffect*> mapPostEffects;
	LPD3DSURF pOrgRenderTarg, pOrgZSurf;		///< original render target saved at BeginScene and restored at EndScene
	D3DVIEWPORT9 d3dView;						///< viewport for rts
	WORD wSceneEffectId;						///< id of the effect that combines all the results

	TCHAR szShadersDir[128];					///< directory that contains shaders folder
	bool bMRTIndepBits;							///< true if D3DPMISCCAPS_INDEPENDENTWRITEMASKS is enabled
	bool bRenderingScene;						///< true if between begin and call blocks
	bool bInteractive;							///< applies changes as they are seen; if false, in developer mode
	bool bAnimSunDir;							///< if true, animate the sun dir

	static SFRenderMngr* pRenderMngr;
	float fTotalElapsed;						///< total elapsed time

	SFLightMngr sflightMngr;
};

inline void SFRenderMngr::SetChannels(DWORD dwChanMask)
{
	//TODO: change the color write bit fields too
	DWORD dwRendTarg = 1;
	while(dwRendTarg < 4) {
		if( dwChanMask & (1<<dwRendTarg) )
			pd3dDevice->SetRenderTarget(dwRendTarg, vecEffectChans[dwRendTarg]->psurf);
		else pd3dDevice->SetRenderTarget(dwRendTarg, NULL);
		++dwRendTarg;
	}
}

#endif